#[allow(unused_must_use)]
pub mod shell {
    use std::process::Command;
    use std::io::stdin;
    use std::io::stdout;
    use std::io::Write;
    use std::path::Path;
    use std::env;
    use std::str;
    use std::process::Stdio;
    use termion::event::Key;
    use termion::input::TermRead;
    use termion::raw::IntoRawMode;
    use termion::raw::RawTerminal;
    use termion::cursor;
    use termion::color;
    use termion;
    use unicode_segmentation::UnicodeSegmentation;

    mod history;
    use crate::shell::history::*;

    mod autocomplete;
    use crate::shell::autocomplete::*;

    mod prompt;
    use crate::shell::prompt::*;

    //used for home directory
    extern crate dirs;



    fn process_line(input: &str) -> String {
        let mut commands = input.trim().split(" | ").peekable();
        let mut previous_command = None;
        let mut resultat = String::new();

        while let Some(command) = commands.next()  {
            let mut parts = command.trim().split_whitespace();
            let command = parts.next().unwrap();
            let args = parts;

            match command {
                "cd" => {
                    let new_dir = args.peekable().peek().map_or("/", |x| *x);
                    let root = Path::new(new_dir);
                    if let Err(e) = env::set_current_dir(&root) {
                        eprintln!(" Err: {}", e);
                    }
                    previous_command = None;
                },
                "history" => {
                    //let stdout = Stdio::inherit();
                    let res = get_history();
                    match res {
                        Ok(r) => return r,
                        Err(e) => eprintln!(" Err: {}", e),
                    }
                },
                "ssh" => {
                    let ssh_args = args.peekable().peek().map_or(" ", |x| *x);
                    let mut child = Command::new("ssh").arg(ssh_args).spawn().unwrap();
                    let _ = child.wait().unwrap();
                },
                command => {
                    if commands.peek().is_some() {
                        let stdin  = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let stdout = Stdio::piped();
                    
                        let output = Command::new(command)
                            .args(args)
                            .stdout(stdout)
                            .stdin(stdin)
                            .spawn();

                        match output {
                            Ok(o) => previous_command = Some(Stdio::from(o.stdout.unwrap())),
                            Err(e) => {
                                previous_command = None;
                                eprintln!(" Err: {}", e);
                            },
                        };
                    } else {
                        let stdin = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };

                        print!("\r\n");
                      
                        let child = match Command::new(command)
                            .args(args)
                            .stdin(stdin)
                            .spawn() {
                                Ok(h) => h,
                                Err(_) => {
                                    let err_string = format!("Command not found : {}\r\n", command);
                                    return String::from(err_string);
                                },
                            };

                        let output = child
                            .wait_with_output()
                            .expect("Failed to wait on child");
                        //RawTerminal::activate_raw_mode(stdout);

//                        let output = Command::new(command)
//                            .args(args)
//                            .stdin(stdin)
//                            .output();

//                        let command_result = match output {
//                            Ok(o) => o,
//                            Err(e) => {
//                                eprintln!("\r\nGot error {}", e);
//                                return String::from("!");
//                            },
//                        };
//
//                        let _ = &mut resultat.push_str(str::from_utf8(&command_result.stdout)
                        let status = output.status
                            .code()
                            .expect("Could not get code")
                            .to_string();
                        previous_command = None;
                        let format_res = format!("{}...Exit: {}\r\n", 
                                                 color::Fg(color::Cyan),
                                                 status);
                        let _ = &mut resultat.push_str(&format_res);
                    }
                },
            };
        }
        return resultat;
    }

    fn replace_signs(line: &String) -> String {
        let mut ayo = String::from(line);
        if ayo.contains('~') {
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            let result = str::replace(&ayo, "~", &homedir);
            ayo = String::from(result);
        }
        
        if ayo.contains("!!") {
            let prev_comm = get_hist_from_number(&get_history_number().unwrap()).unwrap();
            let result = str::replace(&ayo, "!!", &prev_comm);
            ayo = String::from(result);
        }

        return ayo;
    }

    fn handle_input(line: &str) -> String {
        //history callback
        if (line.len() > 0) && (line.starts_with('!'))  {
            let command_found = treat_history_callback(line);
            match command_found {
                Some(c) => {
                    write_to_history(&c);
                    let outp = process_line(&c);
                    return outp;
                    },
                None => return String::from(" "),
            };
        //Regular command
        } else if line.len() > 0 {
            write_to_history(line);
            return process_line(line);
        //Nothing
        } else {
            //Command was empty, new line
            return String::from("");
        }
    }

    fn delete_current_cmd(current_cmd: String) -> String {
        //Deletes currently written command
       
        let mut stdout = stdout().into_raw_mode().unwrap();
        //write!(stdout, "{}", cursor::Save);
        for _c in current_cmd.graphemes(true) {
            write!(stdout, "\x1b[D").unwrap();
            write!(stdout, "\x1b[K").unwrap();
        }
        stdout.flush();
        //write!(stdout, "{}", cursor::Restore);
        return String::new();
    }

    fn get_cmd_curs_pos(command: &String) -> (usize, usize) {
        let mut max_pos: usize = 0; 
        let mut current_pos: usize = 0; 

        for _c in command.graphemes(true) {
            max_pos += 1;
            current_pos += 1;
        }

        return (max_pos, current_pos);
    }

    fn clear_line(max_pos: &usize, current_pos: &usize) {
        //clears currently written command from the screen
        //...NOT from the variable !
        let mut stdout = stdout().into_raw_mode().unwrap();
        let right_spaces: u16 = (*max_pos - *current_pos).try_into().unwrap();
        write!(stdout, "{}", cursor::Right(right_spaces));
        for _i in 0..*max_pos {
            write!(stdout, "{}", cursor::Left(1));
            write!(stdout, "\x1b[K").unwrap();
        }
        write!(stdout, "{}", cursor::Left(1));
        write!(stdout, "\x1b[K").unwrap();
        stdout.flush();
    }



    pub fn run_raw() {
        let stdin = stdin();
        let mut stdout = stdout().into_raw_mode().unwrap();
        //RawTerminal::suspend_raw_mode(&stdout);
        let mut mycommand = String::new();

        //Used in conjunction with Up arrow to go back in history
        //Resetted by pressing Enter
        let mut current_number = get_curr_history_number();
        let mut prompt = build_prompt(&current_number);

        //Used to track the location of the cursor inside the command
        let mut current_pos: usize = 0;
        let mut max_pos: usize = 0;

        //Initialize
        write!(stdout, "\r\n SquiShell (sqish)--- \r\n{}", prompt);
        stdout.flush();

        for c in stdin.keys() {

            match c.unwrap() {
                Key::Char('\t') => {
                    let (res, list) = autocomplete(&mycommand);
                    write!(stdout, "\r\n{}\r\n", list);
                    mycommand.clear();
                    mycommand.push_str(&res);
                    max_pos = res.len();
                    current_pos = max_pos;
                    write!(stdout, "\r\n{}{}", prompt, res);
                    stdout.flush();
                }

                Key::Char('\n') => {
                    current_number = get_curr_history_number();
                    prompt = build_prompt(&current_number);
                    current_pos = 0;
                    max_pos = 0;
                    if (mycommand != String::from("\n")) && (mycommand != String::from("exit")) {
                        let comm = replace_signs(&mycommand);
                        RawTerminal::suspend_raw_mode(&stdout);
                        let res = handle_input(&comm);
                        RawTerminal::activate_raw_mode(&stdout);
                        mycommand.clear();
                        //Proper printing
                        for line in res.split('\n') {
                            if line != "\r" {
                                write!(stdout, "\r\n{}", line);
                            }
                        }
                        current_number = get_curr_history_number();
                        prompt = build_prompt(&current_number);
                        write!(stdout, "{}", prompt).unwrap();
                        stdout.flush();
                    } else if mycommand == String::from("exit") {
                        write!(stdout, "\r\n Sayonara \r\n");
                        break;
                    } else {
                        write!(stdout, "\r\n{}", prompt).unwrap();
                        stdout.flush();
                    }
                },

                Key::Ctrl('q') => {
                    writeln!(stdout, "\r\n Sayonara \r\n");
                    break;
                },

                Key::Ctrl('c') => {
                    current_pos = 0;
                    max_pos = 0;
                    mycommand.clear();
                    write!(stdout, "\r\n--CANCEL--\r\n");
                    current_number = get_curr_history_number();
                    prompt = build_prompt(&current_number);
                    write!(stdout, "{}", prompt);
                },

                Key::Ctrl('t') => {
                    clear_line(&max_pos, &current_pos);
                    max_pos = 0;
                    current_pos = 0;
                    mycommand.clear();
                },

                Key::Ctrl('a') => {
                    if current_pos != 0 {
                        write!(stdout, "{}", cursor::Left(current_pos as u16));
                        current_pos = 0;
                    }
                },

                Key::Ctrl('e') => {
                    if current_pos != max_pos {
                        let chars_left = max_pos - current_pos;
                        write!(stdout, "{}", cursor::Right(chars_left as u16));
                        current_pos = max_pos;
                    }
                },

                Key::Char(c) => {
                    //if at the end of the command...
                    if current_pos == max_pos {
                        mycommand.push(c);
                        current_pos += 1;
                        max_pos += 1;
                        write!(stdout, "{}", c).unwrap();
                    } else if current_pos < max_pos {
                        //Inserting a char inside the command...
                        write!(stdout, "{}", cursor::Save);
                        mycommand.insert(current_pos, c);
                        let command_piece = &mycommand[current_pos..];
                        write!(stdout, "{}", command_piece);
                        stdout.flush();
                        write!(stdout, "{}", cursor::Restore);
                        write!(stdout, "{}", cursor::Right(1));
                        max_pos += 1;
                        current_pos += 1;
                    }

                },

                Key::Backspace => {
                    if current_pos > 0 {
                        current_pos -= 1;
                        max_pos -= 1;
                        mycommand.remove(current_pos);
                        write!(stdout, "{}", cursor::Left(1));
                        write!(stdout, "{}", cursor::Save);
                        clear_line(&max_pos, &current_pos);
                        write!(stdout, "{}", mycommand);
                        write!(stdout, "{}", cursor::Restore);
                        stdout.flush();
                    }
                },

                Key::Up => {
                    mycommand = delete_current_cmd(mycommand);
                    current_number -= 1;
                    let lastcmd = match get_hist_from_number(&current_number) {
                        Some(c) => c,
                        None => continue
                    };
                    mycommand = lastcmd.trim().to_string();
                    (max_pos, current_pos) = get_cmd_curs_pos(&mycommand);
                    //let prompt = build_prompt();
                    write!(stdout, "{}", mycommand);
                },

                Key::Down => {
                    mycommand = delete_current_cmd(mycommand);
                    current_number += 1;
                    let lastcmd = match get_hist_from_number(&current_number) {
                        Some(c) => c,
                        None => continue
                    };
                    mycommand = lastcmd.trim().to_string();
                    (max_pos, current_pos) = get_cmd_curs_pos(&mycommand);
                    //let prompt = build_prompt();
                    write!(stdout, "{}", mycommand);
                },

                Key::Ctrl('p') => {
                    print!("{}", current_pos);
                }

                Key::Ctrl('x') => {
                    clear_line(&max_pos, &current_pos);
                    mycommand.clear();
                    current_pos = 0;
                    max_pos = 0;
                }

                Key::Left => {
                    if current_pos > 0 {
                        current_pos -= 1;
                        write!(stdout, "{}", cursor::Left(1));
                    } 
                },

                Key::Right => {
                    if current_pos < max_pos {
                        print!("{}", cursor::Right(1));
                        current_pos += 1;
                    }
                }

                _ => (),
            }

            stdout.flush().unwrap();
        }

    }


    #[cfg(test)]
    mod tests {
        use super::*;

        fn inittests() {
            let filepath = dirs::home_dir().unwrap().join("history.sqish");

            let file = File::create(&filepath).expect("Could not create test history");
            writeln!(&file, "1 ls");
            writeln!(&file, "2 pwd");
        }

        #[test]
        fn test_gethistnumber() {
            inittests();
            let mynumber = get_history_number().unwrap();
            assert_eq!(mynumber, 2);
        }

        #[test]
        fn test_gethistfromnumber() {
            inittests();
            let mycmd = get_hist_from_number(&1).unwrap();
            assert_eq!(mycmd, String::from("ls"));
        }

        #[test]
        fn test_history_callback() {
            inittests();
            let expected = String::from("ls");
            assert_eq!(treat_history_callback("!1"), Some(expected));
        }

        #[test]
        fn test_replace_signs() {
            inittests();
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            assert_eq!(replace_signs(String::from("ls ~ && echo !!")), String::from(format!("ls {} && echo pwd", homedir)));
        }

        #[test]
        fn test_writehistline() {
            inittests();
            let towrite = String::from("pwd");
            write_to_history(&towrite);

            //Now check directly against the file
            let filepath = dirs::home_dir().unwrap().join("history.sqish");
            let file = File::open(filepath).expect("Error");
            let mut reader = BufReader::new(file).lines();
            let myline = String::from(reader.last().unwrap().expect("Error"));
            let expected = String::from("3 pwd");

            assert_eq!(myline.trim(), expected.trim()); 
        }
    }
}


